package edu.illinois.imunit;

import java.lang.Thread.State;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import edu.illinois.imunit.internal.parsing.BlockEvent;
import edu.illinois.imunit.internal.parsing.Event;
import edu.illinois.imunit.internal.parsing.Name;
import edu.illinois.imunit.internal.parsing.Ordering;
import edu.illinois.imunit.internal.parsing.Orderings;
import edu.illinois.imunit.internal.parsing.SimpleEvent;

/**
 * Provides the API for maintaining schedule state and enforcing schedules.
 * 
 * @author Vilas Jagannath (vbangal2@illinois.edu), <br/>
 *         Milos Gligoric (gliga@illinois.edu), <br/>
 *         Dongyun Jin (djin3@illinois.edu), <br/>
 *         Qingzhou Luo (qluo2@illinois.edu).
 */
public class IMUnit {

    /**
     * Constants
     */
    private static final String AT = "@";
    private static final String CLEARED_SCHEDULE = "";

    /**
     * Current schedule state i.e. state describing the currently enforced schedule.
     */
    private static String currentSchedule = CLEARED_SCHEDULE;
    private static final Map<String, Set<Event>> currentOrderings = new HashMap<String, Set<Event>>();
    private static final Map<String, Thread> currentHappenedEvents = new HashMap<String, Thread>();

    /**
     * Lock used to enforce schedules.
     */
    private static final Object lock = new Object();

    /**
     * Returns the current schedule's name.
     * 
     * @return the current schedule's name.
     */
    public static String getSchedule() {
        return currentSchedule;
    }

    /**
     * Returns <code>true</code> if the given name is the name of the current schedule,
     * <code>false</code> otherwise.
     * 
     * @param name
     * @return <code>true</code> if the given name is the name of the current schedule,
     *         <code>false</code> otherwise.
     */
    public static boolean isSchedule(String name) {
        return currentSchedule.equals(name);
    }

    /**
     * Sets the current schedule state. This method is called by {@link IMUnitRunner} for each
     * test/schedule pair.
     * 
     * @param name
     * @param orderings
     */
    public static void setSchedule(String name, Orderings orderings) {
        /* Note the name of the schedule and the partial orders defined by it */
        currentSchedule = name;
        currentOrderings.clear();
        for (Ordering partialOrder : orderings.getOrderings()) {
            SimpleEvent afterEvent = partialOrder.getAfterEvent();
            String afterEventDesc = getEventDesc(afterEvent);
            Set<Event> beforeEvents = currentOrderings.get(afterEventDesc);
            if (beforeEvents == null) {
                beforeEvents = new HashSet<Event>();
                currentOrderings.put(afterEventDesc, beforeEvents);
            }
            beforeEvents.add(partialOrder.getBeforeEvent());
        }
        currentHappenedEvents.clear();
    }

    /**
     * Clears the current schedule state.
     */
    public static void clearSchedule() {
        currentSchedule = CLEARED_SCHEDULE;
        currentOrderings.clear();
        currentHappenedEvents.clear();
    }

    /**
     * Method called to notify the occurrence of an event. This method contains the logic for
     * enforcing the current schedule. If the fired event needs to occur after some events, the
     * current thread is blocked until those events have occurred.
     * 
     * @param eventName
     */
    public static void fireEvent(String eventName) {
        if (!currentOrderings.isEmpty()) {
            /* Collect the event(s) this event needs to wait for */
            Set<Event> beforeEvents = new HashSet<Event>();
            if (currentOrderings.containsKey(eventName)) {
                beforeEvents.addAll(currentOrderings.get(eventName));
            }
            Thread currentThread = Thread.currentThread();
            String qualifiedName = eventName + AT + currentThread.getName();
            if (currentOrderings.containsKey(qualifiedName)) {
                beforeEvents.addAll(currentOrderings.get(qualifiedName));
            }

            /* Wait for collected events to be completed */
            if (!beforeEvents.isEmpty()) {
                for (Event beforeEvent : beforeEvents) {
                    if (beforeEvent instanceof SimpleEvent) {
                        SimpleEvent simpleEvent = (SimpleEvent) beforeEvent;
                        synchronized (lock) {
                            while (!currentHappenedEvents.containsKey(getEventDesc(simpleEvent))) {
                                try {
                                    lock.wait();
                                } catch (InterruptedException e) {
                                    System.err.println("Interrupted while enforcing schedule!");
                                    e.printStackTrace();
                                    System.exit(2);
                                }
                            }
                        }
                    }
                    else if (beforeEvent instanceof BlockEvent) {
                        BlockEvent blockEvent = (BlockEvent) beforeEvent;
                        String blockEventDesc = getEventDesc(blockEvent);
                        synchronized (lock) {
                            while (!(currentHappenedEvents.containsKey(blockEventDesc) && isBlocked(currentHappenedEvents.get(blockEventDesc)))) {
                                try {
                                    lock.wait(50);
                                } catch (InterruptedException e) {
                                    System.err.println("Interrupted while enforcing schedule!");
                                    e.printStackTrace();
                                    System.exit(2);
                                }
                            }
                        }
                    }
                }
            }
            synchronized (lock) {
                /* This event has now happened */
                currentHappenedEvents.put(eventName, currentThread);
                currentHappenedEvents.put(qualifiedName, currentThread);
                /* Notify other threads that might be waiting for this event to happen */
                lock.notifyAll();
            }
        }
    }

    /**
     * Helper method for constructing the description of a {@link SimpleEvent}.
     * 
     * @param simpleEvent
     * @return description of simpleEvent
     */
    private static String getEventDesc(SimpleEvent simpleEvent) {
        String eventDesc = simpleEvent.getEventName().getName();
        Name eventThreadName = simpleEvent.getThreadName();
        if (eventThreadName.getName() != null) {
            eventDesc += AT + eventThreadName.getName();
        }
        return eventDesc;
    }

    /**
     * Helper method for constructing the description of a {@link BlockEvent}.
     * 
     * @param blockEvent
     * @return description of blockEvent
     */
    private static String getEventDesc(BlockEvent blockEvent) {
        String eventDesc = blockEvent.getBlockAfterEventName().getName();
        Name eventThreadName = blockEvent.getThreadName();
        if (eventThreadName.getName() != null) {
            eventDesc += AT + eventThreadName.getName();
        }
        return eventDesc;
    }

    /**
     * Helper method for checking whether the given {@link Thread} is blocked.
     * 
     * @param thread
     * @return <code>true</code> if thread is blocked and <code>false</code> otherwise.
     */
    private static boolean isBlocked(Thread thread) {
        State state = thread.getState();
        return state.equals(State.BLOCKED) || state.equals(State.WAITING) || state.equals(State.TIMED_WAITING);
    }

}
